﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Windows.Forms;

using Mcs.CMOS.Base;
using Mcs.CMOS.GuiTools;
using Mcs.CMOS.Tools;
using Mcs.CMOS.DeviceSettings;
using Mcs.DataStream;
using Mcs.TChart.FastLine;
using System.Threading.Tasks;
using Mcs.DataTypes;
using Mcs.Hdf5.FileIO;
using System.Linq;

namespace Mcs.CMOS.SpikeView
{
    /// <summary>
    /// CMOS-MEA-Tool to detect and display Spikes.
    /// </summary>
    public class SpikeViewTool// : CMosAnalysisToolBase
    {
        private object selectedSensorsAccess = new object();

        private DataBufferSpike spikes = new DataBufferSpike(new Identifier(), "Spikes");
        private SettingsCMOSMea settingsCMosMea = new SettingsCMOSMea();
        private Dictionary<int, int> selectedSensors = new Dictionary<int, int>();
        private Dictionary<int, Queue<Spike>> accumulatedSpikes = new Dictionary<int, Queue<Spike>>();
        private long overlayInterval = 0;
        private long lastUpdateTime = 0;
        private MapCreator mapCreator;
        private Dictionary<int, RawDataPacket<int>> lastActivityPackets = new Dictionary<int, RawDataPacket<int>>();

        #region properties

        public SpikeExplorerView View { get; private set; }

        public SettingsSpikeViewTool Settings { get; private set; }

        #endregion

        public SpikeViewTool(int id, string name)
        {
            Settings = new SettingsSpikeViewTool();
            Settings.EvtReset += OnSettingsEvtReset; 

            settingsCMosMea.Reset(Settings.ChipType, Settings.ChipType.ToString(), "0001");

            foreach(InstrumentBase.InfoSite info in settingsCMosMea.SensorSiteInfos)
            {
                accumulatedSpikes.Add(info.ID, new Queue<Spike>());
            }

            mapCreator = new MapCreator(name + " MapCreator", false);
            mapCreator.Settings.ActivityType = enActivityType.SpikeCount;
            mapCreator.Settings.AutoMinMax = false;
            mapCreator.Settings.Min = 0;
            mapCreator.Settings.Max = 20;
            mapCreator.Settings.ColorMap = enColorMap.WhiteRed;
            mapCreator.EvtData += OnMapCreatorEvtData;

            View = new SpikeExplorerView();
            View.ToolStripVisible = false;
            //View.EvtControl += OnViewEvtControl;

            View.SensorArray.EvtControl += OnSensorArrayEvtControl;
            View.SensorArray.EvtColorMapMinMax += OnSensorArrayEvtColorMapMinMax;
            View.SensorArray.EvtRoiCursorAdd += OnSensorArrayEvtRoiCursorAdd;
            View.SensorArray.EvtRoiCursorRemove += OnSensorArrayEvtRoiCursorRemove;
            View.SensorArray.EvtRoiCursorPos += OnSensorArrayEvtRoiCursorPos;
            View.SensorArray.EvtRemoveRois += OnSensorArrayEvtRemoveRois;

            View.RasterPlot.Settings.XAxesUnit = new Unit(enUnit.s, 0);
            View.RasterPlot.Settings.Tick = Settings.Tick;
            View.RasterPlot.SetXAxisMinimum(0);
            View.RasterPlot.XSpan(Settings.AccumulateDuration);

            View.Traces.SetTick(Settings.Tick);
            View.Traces.OverlayInterval = 3000;
        }

        private void OnSettingsEvtReset()
        {
            if (settingsCMosMea.Type != Settings.ChipType)
            {
                settingsCMosMea.Reset(Settings.ChipType, Settings.ChipType.ToString(), "0001");
            }
            View.RasterPlot.Settings.Tick = Settings.Tick;
            View.Traces.SetTick(Settings.Tick);
            View.RasterPlot.XSpan(Settings.AccumulateDuration);
        }

        public void Start()
        {
            lastUpdateTime = 0;
            foreach (Queue<Spike> q in accumulatedSpikes.Values)
            {
                q.Clear();
            }
            UpdateSpikesInViews();
        }

        private delegate void DelegateSetCurrentTime(long timeStamp);
        public void SetCurrentTime(long timeStamp)
        {
            if (View.InvokeRequired)
            {
                View.Invoke(new DelegateSetCurrentTime(SetCurrentTime), new object[] { timeStamp });
            }
            else
            {
                foreach (int sensorID in selectedSensors.Values)
                {
                    View.Traces.SetOverlayBufferTimeStamp(sensorID, timeStamp);
                }

                if (timeStamp - lastUpdateTime > Settings.DisplayRefreshTime)
                {
                    long accumulateFrom = 0;
                    if (timeStamp > Settings.AccumulateDuration) // accumulate spikes
                    {
                        accumulateFrom = timeStamp - Settings.AccumulateDuration;
                    }

                    //foreach (int sensorID in accumulatedSpikes.Keys)
                    Parallel.ForEach(accumulatedSpikes.Keys, (sensorID) =>
                    {
                        while (accumulatedSpikes[sensorID].Count > 0)
                        {
                            if (accumulatedSpikes[sensorID].Peek().TimeStamp < accumulateFrom)
                            {
                                accumulatedSpikes[sensorID].Dequeue();
                            }
                            else
                            {
                                break;
                            }
                        }
                    });

                    View.RasterPlot.SetXAxisMinimum(accumulateFrom);
                    UpdateSpikesInViews();
                    lastUpdateTime = timeStamp;
                }
            }
        }

        private delegate void DelegateSetData(Dictionary<int, List<Spike>> spikes);
        public void SetData(Dictionary<int, List<Spike>> spikes)
        {
            if (View.InvokeRequired)
            {
                View.Invoke(new DelegateSetData(SetData), new object[] { spikes });
            }
            else
            {
                UpdateTraceView(spikes);

                Parallel.ForEach(spikes.Keys, (sensorID) =>
                {
                    for (int spike = 0; spike < spikes[sensorID].Count; ++spike)
                    {
                        accumulatedSpikes[sensorID].Enqueue(spikes[sensorID][spike]);
                    }
                });
            }
        }

        #region on view command

        private void OnSensorArrayEvtControl(SensorArrayView.ControlEventArgs obj)
        {
            if (obj.Command == SensorArrayView.enCommand.SelectSensorsAutomatic)
            {
                //ExtractUnits();
            }
            else if (obj.Command == SensorArrayView.enCommand.ExploreOn)
            {
                //View.WaveformPlot.AddPlot(exploreRoiID, "Explore");
            }
            else if (obj.Command == SensorArrayView.enCommand.ExploreOff)
            {
                //View.WaveformPlot.RemovePlot(exploreRoiID);
            }
            else if (obj.Command == SensorArrayView.enCommand.Settings)
            {
                if (obj.Command == SensorArrayView.enCommand.Settings)
                {
                    DlgSettingsSensorArray dlg = new DlgSettingsSensorArray(mapCreator.Settings);
                    dlg.Location = Control.MousePosition;
                    if (dlg.ShowDialog() == DialogResult.OK)
                    {
                    }
                    mapCreator.Work(lastActivityPackets);
                }
            }
        }

        private void OnSensorArrayEvtColorMapMinMax(SensorArrayView.ColorMapMinMaxEventArgs obj)
        {
            mapCreator.Settings.Min = 0;
            mapCreator.Settings.Max = (long) obj.MinMax;
            mapCreator.Work(lastActivityPackets);
        }

        private void OnSensorArrayEvtRoiCursorRemove(MeaField.RoiCursorEventArgs obj)
        {
            int roiID = obj.ID;
            int sensorID = settingsCMosMea.GetSensorInfo(obj.SensorRegion.Left, obj.SensorRegion.Top).ID;
            lock (selectedSensorsAccess)
            {
                if (selectedSensors.Keys.Contains(roiID))
                {
                    selectedSensors.Remove(roiID);
                    RemoveTraces(new List<int>() { roiID });
                }
            }
        }

        private void OnSensorArrayEvtRoiCursorAdd(MeaField.RoiCursorEventArgs obj)
        {
            int roiID = obj.ID;
            int sensorID = settingsCMosMea.GetSensorInfo(obj.SensorRegion.Left, obj.SensorRegion.Top).ID;
            lock (selectedSensorsAccess)
            {
                if (!selectedSensors.Keys.Contains(roiID))
                {
                    if (!selectedSensors.Values.Contains(sensorID))
                    {
                        selectedSensors.Add(roiID, sensorID);
                        AddTraces(new List<int>() { roiID });
                    }
                    else
                    {
                        //View.SensorArray.RemoveRois()
                    }
                }
            }
        }

        private void OnSensorArrayEvtRoiCursorPos(SensorArrayView.RoiCursorPosEventArgs obj)
        {
            if (obj.PreviousSensorRegion.Left != obj.NewSensorRegion.Left || obj.PreviousSensorRegion.Top != obj.NewSensorRegion.Top)
            {
                // remove previous unit
                int roiID = obj.ID;
                int previousSensorID = settingsCMosMea.GetSensorInfo(obj.PreviousSensorRegion.Left, obj.PreviousSensorRegion.Top).ID;
                int sensorID = settingsCMosMea.GetSensorInfo(obj.NewSensorRegion.Left, obj.NewSensorRegion.Top).ID;
                lock (selectedSensorsAccess)
                {
                    if (selectedSensors.Keys.Contains(roiID))
                    {
                        selectedSensors[roiID] = sensorID;
                        AddTraces(new List<int>() { sensorID });
                    }
                }
            }
        }

        private void OnSensorArrayEvtRemoveRois(SettingsRois.RemoveRoisEventArgs obj)
        {
            foreach (int roiID in obj.IDs)
            {
                lock (selectedSensorsAccess)
                {
                    if (selectedSensors.Keys.Contains(roiID))
                    {
                        selectedSensors.Remove(roiID);
                    }
                }
            }
            RemoveTraces(obj.IDs);
        }

        private void AddTraces(List<int> roiIDs)
        {
            Dictionary<int, int> tmp;
            lock (selectedSensorsAccess)
            {
                tmp = new Dictionary<int, int>(selectedSensors);
            }

            List<int> sortedRoiIDs = new List<int>(tmp.Keys);
            sortedRoiIDs.Sort();
            int columns = (int)System.Math.Sqrt(sortedRoiIDs.Count);
            int rows = (int)System.Math.Ceiling((double)sortedRoiIDs.Count / columns);

            View.Traces.SetGridLayout(columns, rows, columns - 1, rows - 1);
            int roi = 0;
            for (int row = 0; row < rows; ++row)
            {
                for (int column = 0; column < columns; ++column)
                {
                    if (roi < sortedRoiIDs.Count)
                    {
                        int roiID = sortedRoiIDs[roi];
                        View.Traces.AddChannel(tmp[roiID], sortedRoiIDs[roi].ToString() + " " + settingsCMosMea.GetSensorInfo(tmp[roiID]).Name, column, row);
                        ++roi;
                    }
                }
            }
        }

        private void RemoveTraces(List<int> sensorIDs)
        {
            Dictionary<int, int> tmp;
            lock (selectedSensorsAccess)
            {
                tmp = new Dictionary<int, int>(selectedSensors);
            }

            List<int> sortedRoiIDs = new List<int>(tmp.Keys);
            sortedRoiIDs.Sort();
            int columns = (int)System.Math.Sqrt(sortedRoiIDs.Count);
            int rows = (int)System.Math.Ceiling((double)sortedRoiIDs.Count / columns);

            View.Traces.SetGridLayout(columns, rows, columns - 1, rows - 1);
            int roi = 0;
            for (int row = 0; row < rows; ++row)
            {
                for (int column = 0; column < columns; ++column)
                {
                    if (roi < sortedRoiIDs.Count)
                    {
                        int roiID = sortedRoiIDs[roi];
                        View.Traces.AddChannel(tmp[roiID], sortedRoiIDs[roi].ToString() + " " + settingsCMosMea.GetSensorInfo(tmp[roiID]).Name, column, row);
                        ++roi;
                    }
                }
            }
        }


        #endregion

        private void OnMapCreatorEvtData(MapCreator.DataEventArgs obj)
        {
            bool forceUpdate = false;
            View.SensorArray.SetMap(obj.Map, forceUpdate);
            if (obj.UpdateColors)
            {
                string labelMin = (obj.MapMin).ToString();
                string labelMax = (obj.MapMax).ToString();
                View.SensorArray.SetLegend(labelMin, labelMax, obj.Colors);
                View.SensorArray.SetToolStripColorMapMax(obj.MapMax);
            }
        }

        private void UpdateSpikesInViews()
        {
            lock (selectedSensorsAccess)
            {
                UpdateSensorArrayView(accumulatedSpikes);
                UpdateRasterPlot(accumulatedSpikes);
            }
        }

        private void UpdateSensorArrayView(Dictionary<int, Queue<Spike>> spikes)
        {
            Frame sensorFrame = settingsCMosMea.SensorFrame;

            int[] spikesPerSensor = new int[sensorFrame.Width*sensorFrame.Height];

            foreach (int sensorID in spikes.Keys)
            {
                int index = settingsCMosMea.GetFrameDataIndex(sensorFrame, sensorID);
                spikesPerSensor[index] = spikes[sensorID].Count;
            }

            DataChunk<int> dataChunk = new DataChunk<int>(0, 0, false, spikesPerSensor);
            Dictionary<int, DataChunk<int>> dataChunks = new Dictionary<int, DataChunk<int>>();
            dataChunks.Add(0, dataChunk);
            RawDataPacket<int> dataPacket = new RawDataPacket<int>(0, 0, dataChunks);

            lastActivityPackets.Clear();
            lastActivityPackets.Add(1, dataPacket);

            mapCreator.ChipRegion = sensorFrame;
            Dictionary<int, InfoFrame> rois = new Dictionary<int, InfoFrame>();
            InfoFrame roi = new InfoFrame();
            roi.ID = 1;
            roi.Frame = sensorFrame;
            rois.Add(roi.ID, roi);
            mapCreator.SensorRois = rois;
            mapCreator.Work(lastActivityPackets);
        }

        private void UpdateRasterPlot(Dictionary<int, Queue<Spike>> spikes)
        {
            int[] dummy = new int[0];
            View.RasterPlot.Traces.RemoveTraces();

            List<int> ids = new List<int>(View.SensorArray.RoisDict.Keys);
            ids.Sort();
            ids.Reverse();

            foreach (int id in ids)
            {
                Roi roi = View.SensorArray.RoisDict[id];
                int sensorID = settingsCMosMea.GetSensorInfo(roi.Left, roi.Top).ID;
                if (spikes.ContainsKey(sensorID))
                {
                    View.RasterPlot.Traces.AddTrace(id);
                    View.RasterPlot.Traces.SetTraceLabel(id, "Roi " + id.ToString());
                    foreach (Spike spike in spikes[sensorID])
                    {
                        View.RasterPlot.AddData(id, spike.TimeStamp, dummy);
                    }
                }
            }
        }


        private void UpdateTraceView(Dictionary<int, List<Spike>> spikes)
        {
            Dictionary<int, int> tmp;
            lock (selectedSensorsAccess)
            {
                tmp = new Dictionary<int, int>(selectedSensors);
            }

            if(overlayInterval != spikes.First().Value.First().Waveform.Length * Settings.Tick)
            {
                overlayInterval = spikes.First().Value.First().Waveform.Length * Settings.Tick;
                View.Traces.OverlayInterval = spikes.First().Value.First().Waveform.Length * Settings.Tick;
            }


            foreach (int sensorID in tmp.Values)
            {
                if (spikes.ContainsKey(sensorID))
                {
                    foreach (Spike spike in spikes[sensorID])
                    {
                        if (spike.Waveform != null && spike.Waveform.Length > 0)
                        {
                            double[] convertedData = new double[spike.Waveform.Length];
                            for (int i = 0; i < spike.Waveform.Length; ++i)
                            {
                                convertedData[i] = spike.Waveform[i];
                            }
                            View.Traces.AddConvertedOverlayData(sensorID, convertedData, spike.TimeStamp);
                        }
                    }
                }
            }
        }


        private delegate void DelegateSetSelectedEvents(ReadOnlyDictionary<int, Mcs.TChart.FastLine.EventMarkerInfo> eventSelection);
        public void SetSelectedEvents(ReadOnlyDictionary<int, Mcs.TChart.FastLine.EventMarkerInfo> eventSelection)
        {
            if (View.InvokeRequired)
            {
                View.Invoke(new DelegateSetSelectedEvents(SetSelectedEvents), eventSelection);
            }
            else
            {
                View.RasterPlot.Markers.RemoveMarkerEntities();

                foreach (Mcs.TChart.FastLine.EventMarkerInfo eventMarkerInfo in eventSelection.Values)
                {
                    View.RasterPlot.Markers.AddNewMarkerEntity(eventMarkerInfo.DisplayMarkerID, new List<int>());
                    View.RasterPlot.Markers.SetMarkerEntityColor(eventMarkerInfo.DisplayMarkerID, eventMarkerInfo.DisplayMarkerColor);
                    View.RasterPlot.Markers.SetMarkerEntityPointer(eventMarkerInfo.DisplayMarkerID, eventMarkerInfo.DisplayPointerStyle, eventMarkerInfo.DisplayPointerSize);
                    View.RasterPlot.Markers.SetMarkerEntityPosition(eventMarkerInfo.DisplayMarkerID, MarkerTool.enMarkerPos.Bottom);

                    //foreach (KeyValuePair<Guid, StreamEvents<uint>> eventStream in data.RawData.Events.EventStreams)
                    //{
                    //    if (eventMarkerInfo.StreamID == eventStream.Key)
                    //    {
                    //        foreach (Events<uint> events in eventStream.Value.Events.Values)
                    //        {
                    //            if (eventMarkerInfo.EventID == events.Info.ID)
                    //            {
                    //                foreach (DataEvent<uint> dataEvent in events.DataEvents)
                    //                {
                    //                    View.RasterPlot.Markers.AddMarker(eventMarkerInfo.DisplayMarkerID, dataEvent.TimeStamp);
                    //                }
                    //            }
                    //        }
                    //    }
                    //}
                }
            }
        }
    }
}
